# -----------------------------
# Example Viewer
# X.Papademetris November 2002
# -----------------------------
#
# The viewer keeps (explicitly) two images in memory
#     currentimage   -- the last loaded image
#     currentresults -- the permuted version of the image
#                               which remaps XY to the appropriate
#                               orientation
#
# ------------------------------------------------------------



package require Itcl


itcl::class MyViewer {
    
    protected common params
    protected variable viewer 0
    protected variable scalewidget 0
    protected variable windowscalewidget 0
    protected variable levelscalewidget 0
    protected variable slicenoscalewidget 0
    protected variable mainwidget 0
    protected variable currentimage 0
    protected variable currentresults 0

    constructor { myname } { 
	set params($this,title) $myname
	$this InitializeParameters }
    destructor { } 

    public method InitializeParameters { }
    public method PermuteImage { } 
    public method FlipScaleAndDisplayImage { }
    public method UpdateDisplay { } 
    public method ResetParametersAndGUI { } 
    public method LoadImage { } 
    public method BuildControlFrameGUI { frame }
    public method GenerateMenuUI { menubar } 
    public method CreateViewer { viewerframe } 
    public method  GenerateUI { basewidget } 
};



# -/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/
# SetDefaults initialize the params array to default values
# The array contains both values and widget names
# -/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/

itcl::body MyViewer::InitializeParameters { } {

    set params($this,slice) XY
    set params($this,flipx) 0
    set params($this,flipy) 1
    set params($this,scale) 2.0
    set params($this,window) 255
    set params($this,level) 128
    set params($this,sliceno) 0
    set params($this,status) "No Image In Memory"

    set mainwidget 0


    set currentimage [ vtkImageData  New ]
    set currentresults [ vtkImageData  New ]
}




# -/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/
# Do Permute takes as input the image stored
# in currentimage and permutes it using
# the flag in params($this,slice). The output is stored
# in currentresults
#
#  currentimage --- params($this,slice) ---> currentresults
#
# -/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/

itcl::body MyViewer::PermuteImage { } {

    set perm [ vtkImagePermute  New ]
    $perm SetInput $currentimage 

    switch $params($this,slice) {
	"XY"  { $perm SetFilteredAxes 0 1 2 }
	"YX"  { $perm SetFilteredAxes 1 0 2 }
	"XZ"  { $perm SetFilteredAxes 0 2 1 } 
	"ZX"  { $perm SetFilteredAxes 2 0 1 } 
	"YZ"  { $perm SetFilteredAxes 1 2 0 } 
	"ZY"  { $perm SetFilteredAxes 2 1 0 } 
    }
    $perm Update
    $currentresults ShallowCopy [ $perm GetOutput ]
    $perm Delete

    set numslices [ lindex [ $currentresults GetDimensions ] 2 ]
    $slicenoscalewidget configure -from 0 -to [ expr $numslices -1 ]
    set params($this,sliceno) [ expr $numslices/2 ]

}
 
# -/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/
#
#  FlipScaleAndDisplayImage takes as input currentresults
#                           and applies the appropriate scaling and
#                           flipping operations. The output of this
#                           pipeline is set as the input of 
#                           the viewer viewer
#                           UpdateDisplay is then called
# -/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/

itcl::body MyViewer::FlipScaleAndDisplayImage { }  {


    # Three Filter Pipeline , flipx,flipy,scale

    set tmp [ vtkImageData  New ]
    $tmp ShallowCopy $currentresults

    if { $params($this,flipx) == 1 } {
	set fx [ vtkImageFlip  New ]
	$fx SetInput $tmp
	$fx SetFilteredAxis  0
	$fx Update
	$tmp ShallowCopy [ $fx GetOutput ]
	$fx Delete
    }

    if { $params($this,flipy) == 1 } {
	set fy [ vtkImageFlip  New ]
	$fy SetInput $tmp
	$fy SetFilteredAxis  1
	$fy Update
	$tmp ShallowCopy [ $fy GetOutput ]
	$fy Delete
    }

    set imageResample [ vtkImageResample  New ]
    $imageResample SetInput $tmp
    $imageResample InterpolateOn
    $imageResample SetAxisMagnificationFactor 0 $params($this,scale) 
    $imageResample SetAxisMagnificationFactor 1 $params($this,scale)
    $imageResample SetAxisMagnificationFactor 2 1.0
    $viewer SetInput [ $imageResample GetOutput ] 
    $imageResample Delete
    $tmp Delete
    UpdateDisplay
}

# -/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/
# UpdateDisplay sets the slice number params($this,sliceno) as well as
#                    the windowing parameters (colormap) of the
#                    viewer stored in viewer
#                    Then it forces a display update through use
#                    of $viewer Render
# -/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/
itcl::body MyViewer::UpdateDisplay { } {

    $viewer SetZSlice [ expr round($params($this,sliceno))]
    $viewer SetColorLevel $params($this,level)
    $viewer SetColorWindow $params($this,window)
    $viewer Render
}

# -/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/
# ResetParametersAndGUI -- resets some value parameters to useful default values
#          -- it is called when a new image is loaded
# -/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/

itcl::body MyViewer::ResetParametersAndGUI { } {


    set params($this,slice) XY
    set params($this,flipx) 0
    set params($this,flipy) 1
    set params($this,scale) 2.0

    set range [ [ [ $currentimage GetPointData ] GetScalars ] GetRange ] 
    set meanrange [ expr 0.5*([ lindex $range 0 ] + [ lindex $range 1 ])]
    set fullrange [ expr [ lindex $range 1 ] - [ lindex $range 0 ]]

    $windowscalewidget configure -from 0 -to $fullrange
    $levelscalewidget configure -from [ lindex $range 0 ] -to [ lindex $range 1 ]

    set params($this,window) $meanrange
    set params($this,level)  $meanrange
    set params($this,sliceno) -1

    set img $currentimage
    set dim [ $img GetDimensions ]
    set vxlsize [ $img GetSpacing ]
    set params($this,status) "Image Dim = $dim Vox= $vxlsize Range = $range"

}
    
# -/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/
# LoadImage is called to load a new image. First the filename is
# obtained, then the image is loaded, and the display is updated!
# -/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/

itcl::body MyViewer::LoadImage { } {
    

    set f  [tk_getOpenFile -title "Load Image" -filetypes { {"VTK Image Format" {.vt}}} ]   
    if { [ string length $f ] < 1 } {
	return 0
    }
    set reader [ vtkStructuredPointsReader  New ]
    $reader SetFileName $f
    $reader Update
    $currentimage ShallowCopy [ $reader GetOutput ]
    $reader Delete

    wm title $mainwidget "Example Viewer : [ file tail $f ]"
    ResetParametersAndGUI
    PermuteImage
    FlipScaleAndDisplayImage
}

# -/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/
# BuildControlFrameGUI is a procedure to generate the graphical
# user interface for the properties window that appears to the left
# of the viewer
# The GUI is build inside the pre-existing ctrlframe widget
#
# This demonstrates the use of 
#    labels, optionmenus, checkbuttons and scale Tk widgets
#
#
# The following variables are connected to widgets
#
#   params($this,slice)   -- controls what kind of slice to display
#   params($this,sliceno) -- actual slice number to display
#   params($this,flipx)   -- 0/1 as to whether to flip the x-axis of the permuted image
#   params($this,flipy)   -- 0/1 as to whether to flip the y-axis of the permuted image
#   params($this,scale)   -- expand the image by this factor
#   params($this,window)  -- sets the window width for the colormap 
#   params($this,level)   -- sets the window level for the colormap 
#
# -/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/

itcl::body MyViewer::BuildControlFrameGUI { ctrlframe } {


    # Slice Selection
    frame $ctrlframe.slice; pack $ctrlframe.slice -side top -expand false -fill x
    label         $ctrlframe.slice.1  -text "Slice:"
    tk_optionMenu $ctrlframe.slice.2 [ itcl::scope  params($this,slice) ] XY YX XZ ZX YZ ZY

    # Add callback to option menu entries
    for { set i 0 } { $i < 6 } { incr i } {
	eval "$ctrlframe.slice.2.menu entryconfigure $i -command { $this PermuteImage; $this FlipScaleAndDisplayImage }"
    }

    pack $ctrlframe.slice.1 $ctrlframe.slice.2 -side left -expand false -fill x 
    # Slice Selection
    frame $ctrlframe.sliceno; pack $ctrlframe.sliceno -side top -expand false -fill x
    label         $ctrlframe.sliceno.1  -text "Slice No:"

    set slicenoscalewidget [ scale $ctrlframe.sliceno.2 -orient horizontal  -from 0 -to 100 -resolution 1 -bigincrement 10 -showvalue 1  -variable [ itcl::scope params($this,sliceno) ] ]

    # Add callback when the mouse is released on the scale widget
    eval "bind $ctrlframe.sliceno.2 <ButtonRelease> { $this UpdateDisplay }"
    pack $ctrlframe.sliceno.1 $ctrlframe.sliceno.2 -side left -expand false -fill x 

     # Flip X and/or Y
    frame $ctrlframe.flip; pack $ctrlframe.flip -side top -expand false -fill x
    eval "checkbutton $ctrlframe.flip.x -variable [ itcl::scope params($this,flipx) ] -text \"Flip X\" -command { $this FlipScaleAndDisplayImage }"
    eval "checkbutton $ctrlframe.flip.y -variable [ itcl::scope params($this,flipy) ] -text \"Flip Y\" -command { $this FlipScaleAndDisplayImage }"
    pack $ctrlframe.flip.x $ctrlframe.flip.y -side left -expand false -fill x

    # Scale Image
    frame $ctrlframe.scale; pack $ctrlframe.scale -side top -expand false -fill x
    label $ctrlframe.scale.la -text "Scale:"
    set scalewidget [ scale $ctrlframe.scale.sc -orient horizontal  -from 0.5 -to 10.0 -resolution 0.25 -bigincrement 1.0 -showvalue 1 -digits 3  -variable [ itcl::scope  params($this,scale) ] ]
    pack $ctrlframe.scale.la $ctrlframe.scale.sc -side left -expand false -fill x
    eval "bind $ctrlframe.scale.sc <ButtonRelease> { $this FlipScaleAndDisplayImage }"

    frame $ctrlframe.window; pack $ctrlframe.window -side top -expand false -fill x
    label $ctrlframe.window.la -text "Window:"
    set windowscalewidget [ scale $ctrlframe.window.sc -orient horizontal  -from 0 -to 256 -resolution 1 -bigincrement 16 -showvalue 1 -digits 3  -variable [ itcl::scope params($this,window) ] ]
    pack $ctrlframe.window.la $ctrlframe.window.sc -side left -expand false -fill x
    eval "bind $ctrlframe.window.sc <ButtonRelease> { $this UpdateDisplay } "

    frame $ctrlframe.level; pack $ctrlframe.level -side top -expand false -fill x
    label $ctrlframe.level.la -text "Level:"
    set levelscalewidget [ scale $ctrlframe.level.sc -orient horizontal  -from 0 -to 256 -resolution 1 -bigincrement 16 -showvalue 1 -digits 3  -variable [ itcl::scope params($this,level) ] ]
    pack $ctrlframe.level.la $ctrlframe.level.sc -side left -expand false -fill x
    eval "bind $ctrlframe.level.sc <ButtonRelease> { $this UpdateDisplay } "

}

# -/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/
# GenerateMenuUI generates a basic menu inside the pre-existing
#   widget menubar
# Two menubuttons are generated and the appropriate callbacks are 
# added
# -/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/

itcl::body MyViewer::GenerateMenuUI { menubar } { 

    $menubar add cascade -underline 0 -label "File"  -menu $menubar.file 
    $menubar add cascade -underline 1 -label "Help"  -menu $menubar.help 

    set filemenu [ menu $menubar.file  ]
    eval "$filemenu add command -label Open -command { $this LoadImage }"
    $filemenu add separator
    $filemenu add command -label "Exit" -command { destroy . ; exit } 

    set helpmenu [ menu $menubar.help  ]

    $helpmenu add command -label "About!" -command {  tk_messageBox -title "About This Application" -type ok -message "Example Image Viewer\n Version 3.0\n X.Papademetris March 2008" }

}

# -/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/
# CreateViewer creates the     vtkTkImageViewerWidget inside 
# a pre-existing frame widget and packs it to the right
# We store the ImageViewer in viewer for easy access
# -/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/
itcl::body MyViewer::CreateViewer { viewerframe } {

    vtkTkImageViewerWidget $viewerframe.right -height 256 -width 256 
    pack $viewerframe.right -side right -expand true -fill both -pady 2
    update
    set viewer [ $viewerframe.right GetImageViewer ]
}

# -/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/
# GenerateUI generates the user interface inside the pre-existing
#   widget basewidget
# Three widgets are packed top to bottom, the menubar,the viewerframe
# and the status text box
# The viewerframe frame is divided into two, the controls are packed
# on the left and the actual image viewer on the right
# -/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/

itcl::body MyViewer::GenerateUI { basewidget } {

    set mainwidget $basewidget
    wm title $mainwidget $params($this,title)
    
    set menubar     [ menu  $basewidget.menu ]
    $basewidget configure -menu $menubar

    set viewerframe [ frame $basewidget.middle -bg black ]
    entry $basewidget.status -textvariable params($this,status) -relief raised -state disabled -width 45 -bg black -fg white

    pack $basewidget.status -side bottom -expand false -fill x -pady 1
    pack $viewerframe -side top -expand true -fill both -padx 2 -pady 2 
    frame $viewerframe.left ; pack $viewerframe.left -side left -padx 2 -pady 2 -fill y -expand f

    GenerateMenuUI $menubar
    BuildControlFrameGUI $viewerframe.left
    CreateViewer $viewerframe
}

# -/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/
# Main Program
#  -/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/

# Hide Default Tk Widget
wm withdraw .
# Create a new toplevel widget and set its dimensions and name
toplevel .base 
wm geometry .base 600x450

# Initialize Variables
set viewer [ MyViewer \#auto "Example Viewer" ]
$viewer GenerateUI .base

# Load an Image
$viewer LoadImage



