#include "vtkLinearICPRegistration.h"
#include "vtkDataSet.h"
#include "vtkLandmarkTransform.h"
#include "vtkObjectFactory.h"
#include "vtkPoints.h"
#include "vtkTransform.h"
#include "vtkPointLocator.h"
#include "vtkMath.h"

vtkCxxRevisionMacro(vtkLinearICPRegistration, "$Revision: 1.0 $");
vtkStandardNewMacro(vtkLinearICPRegistration);

//----------------------------------------------------------------------------

vtkLinearICPRegistration::vtkLinearICPRegistration() : vtkAbstractPointBasedRegistration()
{
  this->StartByMatchingCentroids=1;
  this->TransformationType=2;
  this->OutputTransformation=vtkTransform::New();
}

//----------------------------------------------------------------------------
vtkLinearICPRegistration::~vtkLinearICPRegistration()
{
  this->OutputTransformation->Delete();
}

//----------------------------------------------------------------------------
void vtkLinearICPRegistration::PrintSelf(ostream& os, vtkIndent indent)
{
  this->Superclass::PrintSelf(os,indent);

  os << indent << "OutputTransformation: " << this->OutputTransformation << "\n";
  os << indent << "StartByMatchingCentroids: " << this->StartByMatchingCentroids << "\n";
  os << indent << "TransformationType: " << this->TransformationType << "\n";

}
//----------------------------------------------------------------------------
vtkAbstractTransform* vtkLinearICPRegistration::GetTransformation()
{
  return this->OutputTransformation;
}
//----------------------------------------------------------------------------

int  vtkLinearICPRegistration::Run()
{
  // Check source, target

  if (this->Source == NULL || !this->Source->GetNumberOfPoints())
    {
      vtkErrorMacro(<<"Can't execute with NULL or empty input");
      return 0;
    }

  if (this->Target == NULL || !this->Target->GetNumberOfPoints())
    {
      vtkErrorMacro(<<"Can't execute with NULL or empty target");
      return 0;
    }


  // Create sampled input point set
  vtkPoints* SampledSourcePoints=this->SampleDataSet(this->Source,this->MaximumNumberOfPoints);

  // Create locator
  vtkPointLocator* Locator = vtkPointLocator::New();
  Locator->SetDataSet(this->Target);
  //  Locator->SetNumberOfCellsPerBucket(1);
  Locator->BuildLocator();

  // Get The Landmark Transform All Set
  vtkLandmarkTransform* LandmarkTransform=vtkLandmarkTransform::New();

  switch(this->TransformationType)
    {
    case 0:
      LandmarkTransform->SetModeToRigidBody();
      break;
    case 1:
      LandmarkTransform->SetModeToSimilarity();
      break;
    case 2:
      LandmarkTransform->SetModeToAffine();
      break;
    }
  this->OutputTransformation->Identity();
  this->OutputTransformation->PostMultiply();

  // Get The Point Sets Ready
  vtkPoints* CorrespondingPoints=vtkPoints::New();
  CorrespondingPoints->SetNumberOfPoints(SampledSourcePoints->GetNumberOfPoints());

  double offset[3] = { 0.0,0.0,0.0};
  if (this->StartByMatchingCentroids)
    {
      double source_centroid[3]; this->GetCentroid(this->Source,source_centroid);
      double target_centroid[3]; this->GetCentroid(this->Target,target_centroid);
      for (int ia=0;ia<=2;ia++)
	offset[ia]=target_centroid[ia]-source_centroid[ia];

      this->ShiftPoints(SampledSourcePoints,offset,1.0);
    }

  // Begin Alternating Estimation
  int     NumberOfLandmarks=SampledSourcePoints->GetNumberOfPoints();
  int     NumberOfIterations = 1;

  // Provide Feedback to GUI
  // -----------------------
  this->UpdateProgress(0.0);

  double previousmaxdist=0.0;

  while (NumberOfIterations <= this->MaximumNumberOfIterations)
    {

      // Two step procedure
      // First Find Corresponding Points
      // Then estimate transformation!

      // 1. Find Correspondences
      // -----------------------
      for(int i = 0; i < NumberOfLandmarks;i++)
	{
	  // Get a Point
	  double x[3]; SampledSourcePoints->GetPoint(i,x);

	  // Find where this point was at the last estimate!
	  double tx[3]; LandmarkTransform->TransformPoint(x,tx);

	  // Find Corresponding point
	  int id=Locator->FindClosestPoint(tx);
	  CorrespondingPoints->SetPoint(i,this->Target->GetPoint(id));
	}

      // 2. Estimate Transformation
      // --------------------------
      LandmarkTransform->SetSourceLandmarks(SampledSourcePoints);
      LandmarkTransform->SetTargetLandmarks(CorrespondingPoints);
      LandmarkTransform->Update();

      // 3. Generate Output (including offset shift)
      // -------------------------------------------
      this->OutputTransformation->Identity();
      this->OutputTransformation->Translate(offset);
      this->OutputTransformation->Concatenate(LandmarkTransform);

      double progress=double(NumberOfIterations)/double(this->MaximumNumberOfIterations);
      this->UpdateProgress(progress);


      // 4. Check Convergence
      // -------------------------------------------
      double maxdist=0.0;
      for(int i = 0; i < NumberOfLandmarks;i++)
	{
	  // Get a Point
	  double x[3];  SampledSourcePoints->GetPoint(i,x);
	  double tx[3]; LandmarkTransform->TransformPoint(x,tx);
	  double y[3];  CorrespondingPoints->GetPoint(i,y);

	  double dist=vtkMath::Distance2BetweenPoints(tx,y);
	  if (dist>maxdist)
	    maxdist=dist;
	}
      maxdist=sqrt(maxdist);

      if (NumberOfIterations>2)
	{
	  if (fabs(maxdist-previousmaxdist)<this->Epsilon)
	    NumberOfIterations=this->MaximumNumberOfIterations+1;
	}
      previousmaxdist=maxdist;

      vtkDebugMacro(<<"End of Iteration " << NumberOfIterations << "/" << this->MaximumNumberOfIterations<<"\n");
      ++NumberOfIterations;

    }

  Locator->Delete();
  SampledSourcePoints->Delete();
  CorrespondingPoints->Delete();
  LandmarkTransform->Delete();
  this->UpdateProgress(1.0);
  return 1;
}
