#include "vtkObjectFactory.h"
#include "vtkKMeansSegmentation.h"
#include "vtkDoubleArray.h"
#include "vtkImageAccumulate.h"
#include "vtkMath.h"
#include "vtkPointData.h"
#include "vtkInformation.h"
#include "vtkInformationVector.h"
#include "vtkStreamingDemandDrivenPipeline.h"

//------------------------------------------------------------------------------

vtkKMeansSegmentation* vtkKMeansSegmentation::New()
{
  // First try to create the object from the vtkObjectFactory
  vtkObject* ret = vtkObjectFactory::CreateInstance("vtkKMeansSegmentation");
  if(ret)
    {
      return (vtkKMeansSegmentation*)ret;
    }
  // If the factory was unable to create the object, then create it here.
  return new vtkKMeansSegmentation;
}

// Construct object with no children.
vtkKMeansSegmentation::vtkKMeansSegmentation()
{
  this->NumberOfClasses=3;
  this->Iterations=30;
  this->Parameters=NULL;
  this->Histogram=NULL;
  this->Convergence=0.5;
  this->NumberOfBins=128;
  this->HistogramOrigin=0.0;
  this->HistogramSpacing=1.0;
  this->MaxSigmaRatio=0.2;
}
// ----------------------------------------------------------------------------
vtkKMeansSegmentation::~vtkKMeansSegmentation()
{
  
  if (this->Parameters!=NULL)
    this->Parameters->Delete();
  this->Parameters=NULL;

  if (this->Histogram!=NULL)
    this->Histogram->Delete();
  this->Histogram=NULL;



}
// ----------------------------------------------------------------------------
double vtkKMeansSegmentation::Metric(double x,double m,double sigma2)
{
  double t= (x-m)*(x-m)/(-2.0*sigma2);
  double A= 1.0/sqrt(2.0*vtkMath::Pi()*sigma2);
  return A*exp(t);
}
// ----------------------------------------------------------------------------
vtkImageData* vtkKMeansSegmentation::CreateHistogram(vtkImageData* input,int NumBins)
{
  if (input==NULL)
    return NULL;
  
  if (NumBins<4)
    NumBins=4;

  double range[2]; input->GetPointData()->GetScalars()->GetRange(range);
  double minv=range[0];
  double maxv=range[1];

  this->HistogramOrigin=minv;
  this->HistogramSpacing=1.0;

  this->NumberOfBins=NumBins;
  int drange=int(maxv-minv+1);
  
  if (drange< this->NumberOfBins)
    this->NumberOfBins=drange;
  
  while(drange> this->NumberOfBins*this->HistogramSpacing)
    this->HistogramSpacing+=1.0;
  
  this->NumberOfBins=int(drange/this->HistogramSpacing+0.5);

  vtkImageAccumulate* accumulate=vtkImageAccumulate::New();
  accumulate->SetInput(input);
  accumulate->SetComponentOrigin(this->HistogramOrigin,0.0,0.0);
  accumulate->SetComponentSpacing(this->HistogramSpacing,1.0,1.0);
  accumulate->SetComponentExtent(0,this->NumberOfBins-1,0,0,0,0);
  accumulate->Update();

  vtkImageData* out=vtkImageData::New();
  out->ShallowCopy(accumulate->GetOutput());
  accumulate->Delete();
  return out;
}
// --------------------------------------------------------------------------------------------

int vtkKMeansSegmentation::InitializeParameters(vtkImageData* histogram,int NumClasses,vtkDoubleArray* params)
{
  if (histogram==NULL)
    return 0;

  this->NumberOfClasses=NumClasses;
  if (this->NumberOfClasses<1)
    this->NumberOfClasses=1;
  
  params->SetNumberOfComponents(3);
  params->SetNumberOfTuples(this->NumberOfClasses);
  params->FillComponent(0,1.0);

  int dim[3]; this->Histogram->GetDimensions(dim);

  double range[2];
  range[0]=this->HistogramOrigin;
  range[1]=this->HistogramSpacing*double(dim[0]-1);

  int numtuples=params->GetNumberOfTuples();
  double s=(range[1]-range[0])/double(this->NumberOfClasses);

  for (int j=0;j<numtuples;j++)
    {
      double s=(range[1]-range[0])/double(this->NumberOfClasses);
      double m =range[0]+ double(j+0.5)/double(this->NumberOfClasses)*(range[1]-range[0]);
      params->SetComponent(j,1,m);
      params->SetComponent(j,2,s);
      params->SetComponent(j,0,1.0);
    }
 

  return 1;
}


// ---------------------------------------------------------------------------------------------
//
//     Execute Segmentation 
//
// ----------------------------------------------------------------------------
int vtkKMeansSegmentation::RequestInformation(vtkInformation *request,  vtkInformationVector **inputVector, 
					   vtkInformationVector *outputVector)
{
	
	vtkInformation *outInfo = outputVector->GetInformationObject(0);
	vtkDataObject::SetPointDataActiveScalarInfo(outputVector->GetInformationObject(0), -1, 1);
	vtkDataObject::SetPointDataActiveScalarInfo(outInfo, VTK_SHORT, -1);

  return 1;
}

void vtkKMeansSegmentation::SimpleExecute(vtkImageData* input,vtkImageData* output )
{
  if (input==NULL)
    {
      vtkErrorMacro(<<"Bad Inputs to vtkKMeansSegmentation");
      return;
    }
  
  if (this->Histogram!=NULL)
    this->Histogram->Delete();
  this->Histogram=this->CreateHistogram(input,this->NumberOfBins);

  if (this->Parameters!=NULL)
    this->Parameters->Delete();
  this->Parameters=vtkDoubleArray::New();
  this->InitializeParameters(this->Histogram,this->NumberOfClasses,this->Parameters);

  this->UpdateProgress(0.05);

  int dim[3]; this->Histogram->GetDimensions(dim);
  vtkDataArray* data=this->Histogram->GetPointData()->GetScalars();

  // ----------------- EM Estimation -------------------------------------

  double* mean =new double[this->NumberOfClasses];
  double* sigma2=new double[this->NumberOfClasses];
  double* sum  =new double[this->NumberOfClasses];
  double* sum2 =new double[this->NumberOfClasses];
  double*  num =new double[this->NumberOfClasses];

  fprintf(stderr,"\n**** Initial Parameters (maxsigmaratio=%.3f) \n --------------------------\n",
	  this->MaxSigmaRatio);
  for (int j=0;j<this->NumberOfClasses;j++)
    {
      sigma2[j]=this->Parameters->GetComponent(j,2);      sigma2[j]*=sigma2[j];
      mean[j] =this->Parameters->GetComponent(j,1);
      fprintf(stderr,"\t c=%d   mean=%.1f   sigma=%.1f\n",j+1,mean[j],sqrt(sigma2[j]));
    }

  // ----------------------------------------------------------------------------------
  //            Iterations
  // ----------------------------------------------------------------------------------
  int iter=1;
  double error=this->Convergence+1.0;
  
  while (iter <= this->Iterations && error > this->Convergence)
    {
      error=0.0;
      double totalnum=0.0;
      
      for (int i=0;i<this->NumberOfClasses;i++)	{ sum[i]=0.0; sum2[i]=0.0; num[i]=0.0; }
      
      for (int bin=0;bin<dim[0];bin++)
	{
	  double v=this->HistogramOrigin+double(bin)*this->HistogramSpacing;
	  double numv=data->GetComponent(bin,0);
	  double bestp=0.0;
	  int   bestc=0;

	  for (int c=0;c<this->NumberOfClasses;c++)
	    {
	      double p=this->Metric(v,mean[c],sigma2[c]);
	      if (p>bestp)  
		{  
		  bestp=p;  
		  bestc=c;  
		}
	    }
	  num[bestc]+=numv;  
	  sum[bestc]+=v*numv;  
	  sum2[bestc]+=v*v*numv;  
	  totalnum+=numv;
	}
      
     fprintf(stderr,"It=%03d : \t",iter);
     
     // Compute Parameters
     // ------------------
     double maxsigma=0.0;
     for (int c=0;c<this->NumberOfClasses;c++)
       {
	 if (num[c]<1.0)  
	   num[c]=1.0;

	 double m=sum[c]/num[c];
	 if (fabs(m-mean[c])>error)
	   error=fabs(m-mean[c]);
	 
	 mean[c]=m;
	 sigma2[c]=(sum2[c]/num[c]-mean[c]*mean[c]);

	 double s=sqrt(sigma2[c]);
	 if (s>maxsigma)
	   maxsigma=s;
       }

     // Enforce MaxSigmaRatio (0.1)
     // ---------------------------
     for (int c=0;c<this->NumberOfClasses;c++)
       {
	 double s=sqrt(sigma2[c]);
	 if (s<this->MaxSigmaRatio*maxsigma)
	   sigma2[c]=pow(this->MaxSigmaRatio*maxsigma,double(2.0));
	 this->Parameters->SetComponent(c,1,mean[c]);
	 this->Parameters->SetComponent(c,2,sqrt(sigma2[c]));
	 this->Parameters->SetComponent(c,0,num[c]);
	 fprintf(stderr,"%.1f + %.1f (%5.2f)\t",mean[c],sqrt(sigma2[c]),100.0*num[c]/totalnum);
       }
      fprintf(stderr,"error=%.2f\n",error);
      ++iter;
    }

  // At this point sort via mean[c]
  // ------------------------------
  fprintf(stderr,"\n\nSorting \n-----------------\n");
  for (int c=0;c<this->NumberOfClasses;c++)
    {
      int best=c;
      double bestm=mean[c];

      for (int c1=0;c1<this->NumberOfClasses;c1++)
	{
	  if (mean[c1]<bestm)
	    {
	      bestm=mean[c1];
	      best=c1;
	    }
	}
      
      this->Parameters->SetComponent(c,1,mean[best]);
      this->Parameters->SetComponent(c,2,sqrt(sigma2[best]));
      this->Parameters->SetComponent(c,0,num[best]);
      mean[best]=1e+10;

      fprintf(stderr,"Class %d : orig=%d mean=%.2f sigma=%.2f number=%.2f\n",c,best,
	      this->Parameters->GetComponent(c,1),
	      this->Parameters->GetComponent(c,2),this->Parameters->GetComponent(c,0));



    }

  for (int c=0;c<this->NumberOfClasses;c++)
    {
      mean[c]=this->Parameters->GetComponent(c,1);
      sigma2[c]=pow(this->Parameters->GetComponent(c,2),2.0);
    }


  vtkDataArray* inpdata=input->GetPointData()->GetScalars();
  vtkDataArray* outdata=output->GetPointData()->GetScalars();
  outdata->FillComponent(0,-1.0);
  int numvoxels=inpdata->GetNumberOfTuples();

  int index=29+31*51+26*(51*62);
  for (int voxel=0;voxel<numvoxels;voxel++)
    {
      double bestp=0.0;
      int    bestc=0;
      double v=inpdata->GetComponent(voxel,0);
      
      for (int c=0;c<this->NumberOfClasses;c++)
	{
	  double p=this->Metric(v,mean[c],sigma2[c]);
	  if (voxel==index)
	    fprintf(stderr,"c=%d mean[c]=%.2f sigma2[c]=%.2f  \t\t v=%.2f p=%.4f\n",c,mean[c],sigma2[c],v,p);
	  if (p>bestp || c==0)
	    {  
	      bestp=p;  
	      bestc=c; 
	    }
	}
      if (voxel==index)
	fprintf(stderr,"Bestc = %d bestp=%.5f\n",bestc,bestp);
      outdata->SetComponent(voxel,0,bestc);
    }

  delete [] mean;  delete [] sigma2;  delete [] sum;  delete [] sum2;  delete [] num;
}

