#include "vtkImageImport.h"
#include "vtkImageExport.h"
#include "vtkImageData.h"
#include "itkImage.h"
#include "itkImageToImageFilter.h"
#include "itkCurvatureAnisotropicDiffusionImageFilter.h"
#include "itkImageFileReader.h"
#include "itkImageFileWriter.h"
#include "vtkObjectFactory.h"
#include "itkVTKImageExport.h"
#include "itkVTKImageImport.h"
#include "vtkITKUtility.h"
#include "vtkITKMyUtility.h"
#include "vtkImageCast.h"

// This creates the static member function New
vtkStandardNewMacro(vtkITKMyUtility);


// ------------------------------------------------------------------------------------------------------
//
//  Part I -- Using the ITK Curvature Flow Image Filter
//
// ------------------------------------------------------------------------------------------------------
template <class IT>
void vtkITKMyUtilityCurvatureAnisotropicDiffusionSmoothImage(vtkImageData* input,
							     vtkImageData* output,
							     double Conductance,
							     double TimeStep,
							     int NumberOfIterations)
{
  // Define the parts of the ITK pipeline.
  typedef itk::Image<IT, 3> ImageType;
  typedef itk::VTKImageImport<ImageType> ImportType;
  typedef itk::VTKImageExport<ImageType> ExportType;
  typedef itk::CurvatureAnisotropicDiffusionImageFilter<ImageType, ImageType >  FilterType;



  // From VTK to ITK
  //------------------------------------------------------------------------
  vtkImageExport *vtkInputExporter = vtkImageExport::New();
  vtkInputExporter->SetInput(input);
  typename ImportType::Pointer itkInputImporter =  ImportType::New();
  ConnectPipelines(vtkInputExporter, itkInputImporter);
  itkInputImporter->Update();

  // Create the itk::CurvatureAnisotropicDiffusionImageFilter and connect it
  // -----------------------------------------------------------------------
  typename FilterType::Pointer filter = FilterType::New();
  filter->SetInput(itkInputImporter->GetOutput());
  filter->SetTimeStep(TimeStep);
  filter->SetNumberOfIterations(NumberOfIterations);
  filter->SetConductanceParameter( Conductance );
  filter->Update();

  // Back to VTK
  //------------------------------------------------------------------------
  typename ExportType::Pointer itkExporter = ExportType::New();
  itkExporter->SetInput(filter->GetOutput());
  itkExporter->Update();

  vtkImageImport* vtkImporter=vtkImageImport::New();
  ConnectPipelines(itkExporter, vtkImporter);
  vtkImporter->Update();

  output->DeepCopy(vtkImporter->GetOutput());

  vtkImporter->Delete();
  vtkInputExporter->Delete();
  
}

// ------------------------------------------------------------------------------------------------------------------

vtkImageData* vtkITKMyUtility::CurvatureAnisotropicDiffusionFilter(vtkImageData* input,double Conductance,
								 double TimeStep,int NumberOfIterations)
{
  if (input==NULL)
    {
      return NULL;
    }

  vtkImageData* output=vtkImageData::New();
  output->CopyStructure(input);


  if (input->GetScalarType() == VTK_FLOAT )
    {
      vtkITKMyUtilityCurvatureAnisotropicDiffusionSmoothImage<float>(input,output,Conductance,TimeStep,NumberOfIterations);
    } 
  else if (input->GetScalarType() == VTK_DOUBLE )
    {
      vtkITKMyUtilityCurvatureAnisotropicDiffusionSmoothImage<double>(input,output,Conductance,TimeStep,NumberOfIterations);
    }
  else
    {
      vtkImageCast* cst=vtkImageCast::New();
      cst->SetInput(input);
      cst->SetOutputScalarTypeToFloat();
      cst->Update();
      vtkITKMyUtilityCurvatureAnisotropicDiffusionSmoothImage<float>(cst->GetOutput(),output,Conductance,TimeStep,NumberOfIterations);
      cst->Delete();
    }

  return output;
}

// ------------------------------------------------------------------------------------------------------
//
//  Part II -- Using the ITK Image Load Factory
//
// ------------------------------------------------------------------------------------------------------
template <class IT>
vtkImageData* vtkITKMyUtilityLoadImage(const char* fname)
{
  typedef itk::Image< IT,  3 >    ImageType;
  typedef itk::VTKImageImport<ImageType> ImportType;
  typedef itk::VTKImageExport<ImageType> ExportType;
  typedef itk::ImageFileReader< ImageType  >  ReaderType;

  typename ReaderType::Pointer reader = ReaderType::New();
  reader->SetFileName(fname);

  try    {    
    reader->Update();    
  }  
  catch (itk::ExceptionObject & e)    
    {    
      std::cerr << "Failed to read " << fname << "\n" << e.GetDescription() << std::endl;    
      return NULL;    
    }

  typename ExportType::Pointer itkExporter = ExportType::New();
  itkExporter->SetInput(reader->GetOutput());
  itkExporter->Update();

  vtkImageImport* vtkImporter=vtkImageImport::New();
  ConnectPipelines(itkExporter, vtkImporter);
  vtkImporter->Update();

  vtkImageData* output=vtkImageData::New();
  output->DeepCopy(vtkImporter->GetOutput());
  vtkImporter->Delete();

  return output;
}  



vtkImageData* vtkITKMyUtility::LoadImage(const char* filename)
{

  // 
  //  Some of this code derives from code posted by Hideaki Hiraki
  //  on the Insight-users mailing list
  //

  itk::ImageIOBase::Pointer imageIO;
  imageIO=itk::ImageIOFactory::CreateImageIO(filename, itk::ImageIOFactory::ReadMode);

  if (imageIO==static_cast<itk::ImageIOBase::Pointer>(0))
    {
      fprintf(stderr,"Failed to Create Factory\n");
      return NULL;
    }

  imageIO->SetFileName(filename);
  imageIO->ReadImageInformation();
 
  switch( imageIO->GetComponentType() ){
  case itk::ImageIOBase::UCHAR:                
    return vtkITKMyUtilityLoadImage<unsigned char>(filename);
    break;
  case itk::ImageIOBase::CHAR:                 
    return vtkITKMyUtilityLoadImage<char>(filename); 
    break;
  case itk::ImageIOBase::USHORT:                
    return vtkITKMyUtilityLoadImage<unsigned short>(filename); 
    break;
  case itk::ImageIOBase::SHORT: 
    return vtkITKMyUtilityLoadImage<short>(filename); 
    break;
  case itk::ImageIOBase::UINT:
    return vtkITKMyUtilityLoadImage<unsigned int>(filename); 
    break;
  case itk::ImageIOBase::INT: 
    return vtkITKMyUtilityLoadImage<int>(filename); 
    break;
  case itk::ImageIOBase::ULONG: 
    return vtkITKMyUtilityLoadImage<unsigned long>(filename); 
    break;
  case itk::ImageIOBase::LONG:
    return vtkITKMyUtilityLoadImage<long>(filename); 
    break; 
  case itk::ImageIOBase::FLOAT:
    return vtkITKMyUtilityLoadImage<float>(filename); 
    break;
  case itk::ImageIOBase::DOUBLE:
    return vtkITKMyUtilityLoadImage<double>(filename); 
    break;
  case itk::ImageIOBase::UNKNOWNCOMPONENTTYPE:
    fprintf(stderr,"Unknown Type in vtkITKMyUtility::LoadImage");
    break;
  }

  return NULL;
}

// ------------------------------------------------------------------------------------------------------
//
//  Part III  -- Using the ITK Image Save Factory
//
// ------------------------------------------------------------------------------------------------------
template <class IT,int dimensions>
int vtkITKMyUtilitySaveImage(vtkImageData* img,const char* fname,int& flag)
{
  typedef itk::Image< IT,  dimensions >    ImageType;
  typedef itk::ImageFileWriter< ImageType  >  WriterType;
  typedef itk::VTKImageImport<ImageType> ImportType;
  typedef itk::VTKImageExport<ImageType> ExportType;


  // From VTK to ITK
  //------------------------------------------------------------------------
  vtkImageExport *vtkInputExporter = vtkImageExport::New();
  vtkInputExporter->SetInput(img);
  typename ImportType::Pointer itkInputImporter =  ImportType::New();
  ConnectPipelines(vtkInputExporter, itkInputImporter);
  itkInputImporter->Update();
  
  
  // Save Image
  //------------------------------------------------------------------------
  typename WriterType::Pointer writer = WriterType::New();
  writer->SetInput(itkInputImporter->GetOutput());
  writer->SetFileName(fname);
  flag=1;

  try    {    
    writer->Update();    
  }  
  catch (itk::ExceptionObject & e)    
    {    
      std::cerr << "Failed to read " << fname << "\n" << e.GetDescription() << std::endl;    
      flag=0;    
    }

  vtkInputExporter->Delete();
  return flag;
}  

template <class IT>
int vtkITKMyUtilitySaveImage1(vtkImageData* img,const char* fname,int dimension,int& flag)
{
  if (dimension==2)
    return vtkITKMyUtilitySaveImage<IT,2>(img,fname,flag);

  return vtkITKMyUtilitySaveImage<IT,3>(img,fname,flag);
}



int vtkITKMyUtility::SaveImage(vtkImageData* input,const char* filename)
{
  int dimension=3;
  int dim[3]; input->GetDimensions(dim);
  if (dim[2]==1)
    dimension=2;


  int flag=0;

 switch (input->GetScalarType())
    {
    case VTK_FLOAT : 
      flag=vtkITKMyUtilitySaveImage1<float>(input,filename,dimension,flag);
      break;
    case VTK_DOUBLE : 
      flag=vtkITKMyUtilitySaveImage1<double>(input,filename,dimension,flag);
      break;
    case VTK_SHORT : 
      flag=vtkITKMyUtilitySaveImage1<short>(input,filename,dimension,flag);
    case VTK_UNSIGNED_SHORT : 
      flag=vtkITKMyUtilitySaveImage1<unsigned short>(input,filename,dimension,flag);
    case VTK_INT : 
      flag=vtkITKMyUtilitySaveImage1<int>(input,filename,dimension,flag);
    case VTK_UNSIGNED_INT : 
      flag=vtkITKMyUtilitySaveImage1<unsigned int>(input,filename,dimension,flag);
    case VTK_UNSIGNED_CHAR : 
      flag=vtkITKMyUtilitySaveImage1<unsigned char>(input,filename,dimension,flag);
    case VTK_CHAR : 
      flag=vtkITKMyUtilitySaveImage1<char>(input,filename,dimension,flag);
    default:
      fprintf(stderr,"Unsupported Image Type\n");
      flag=0;
      break;
    }

  return flag;

}





